from django.conf import settings
from rest_framework import throttling
from django.core.exceptions import ImproperlyConfigured


class SecurityRateThrottle(throttling.SimpleRateThrottle):
    """Обязательная бликировка после каждого запроса"""

    security_throttle_anon = 'security_throttle_anon'
    security_throttle_user = 'security_throttle_user'

    def __init__(self):
        self._anon_rate = 5  # Количество секунд на 1 запрос для анонимных
        self._user_rate = 10  # Количество секунд на 1 запрос для авторизованных
        self._rate = None

    def get_rate(self):
        raise NotImplemented

    @property
    def rate(self):
        if not getattr(self, '_rate', None):
            msg = ("You must set either `.rate` for '%s' throttle" %
                   self.__class__.__name__)
            raise ImproperlyConfigured(msg)
        return int(self._rate)

    def set_rate(self, request, view):
        if request.user.is_authenticated:
            view_rate = getattr(view, self.security_throttle_user, None)
            self._rate = view_rate if view_rate and view_rate >= 1 else self._user_rate
        else:
            view_rate = getattr(view, self.security_throttle_anon, None)
            self._rate = view_rate if view_rate and view_rate >= 1 else self._anon_rate

    def parse_rate(self, rate):
        return 1, rate

    def get_cache_key(self, request, view):
        try:
            scope = f'{view.__class__.__name__}_{view.action}'
        except AttributeError:
            scope = f'{view.__class__.__name__}'

        if request.user.is_authenticated:
            ident = request.user.pk
        else:
            ident = self.get_ident(request)

        return self.cache_format % {
            'scope': scope,
            'ident': ident
        }

    def allow_request(self, request, view):
        if settings.DEBUG:
            return True
        self.set_rate(request, view)
        rate = self.rate
        self.num_requests, self.duration = self.parse_rate(rate)
        return super().allow_request(request, view)



import datetime
import io
import json
import os

import requests
from google.cloud.exceptions import NotFound
from google.protobuf import timestamp_pb2
from google.cloud import pubsub_v1, storage
from google.cloud import tasks_v2
from google.cloud import error_reporting


report = error_reporting.Client()


class PubSub:
    """ Google Cloud PubSub Adapter"""

    __publisher = pubsub_v1.PublisherClient()

    def __init__(self, topic_id: str, access_token: str = None, project='expressmoney'):
        self.__topic_path = self.__publisher.topic_path(project, topic_id)
        self.__access_token = str(access_token)

    def publish(self, payload: dict = None):
        attrs = {'topic': self.__topic_path, }

        if payload and not isinstance(payload, dict):
            raise TypeError('Payload should be dict type')

        if payload:
            if self.__access_token:
                payload.update({'access_token': self.__access_token})
        else:
            if self.__access_token:
                payload = {'access_token': self.__access_token}
        if payload:
            attrs.update({'data': json.dumps(payload, ensure_ascii=False).encode('UTF-8')})

        self.__publisher.publish(**attrs)


class Tasks:
    """ Google Cloud Tasks Adapter"""

    _client = tasks_v2.CloudTasksClient()

    def __init__(self,
                 access_token: str = None,
                 queue: str = 'attempts-1',
                 location: str = 'europe-west1',
                 project: str = 'expressmoney',
                 in_seconds: int = None,
                 ):
        """
        CloudTasks client
        Args:
            access_token: 'Bearer DFD4345345D'
            queue: 'my-appengine-queue'
            location: 'europe-west1'
            project: 'expressmoney'
            in_seconds: None
        """

        self._project = project
        self._in_seconds = in_seconds
        self._parent = self._client.queue_path(self._project, location, queue)
        self._uri = None
        self._payload = None
        self._update = False
        self._service = None
        self._access_token = access_token

    def run(self,
            service: str,
            uri: str,
            payload: dict = None,
            update: bool = False,
            ):
        """
        Start service execution
        Args:
            service: service name of App Engine
            uri: '/example_task_handler'
            payload: {'param': 'value'}
            update: True - PUT method
        """
        self._service = service
        self._uri = uri
        self._payload = payload
        self._update = update
        task = self._create_task()
        task = self._add_payload(task)
        task = self._convert_in_seconds(task)
        task = self._add_authorization(task)
        task = self._remove_empty_headers(task)
        self._client.create_task(parent=self._parent, task=task)

    def _create_task(self):
        task = {
            'app_engine_http_request': {
                'http_method': self._http_method,
                'relative_uri': self._uri,
                'headers': {},
                'app_engine_routing': {
                    'service': self._service,
                    'version': '',
                    'instance': '',
                    'host': '',

                }
            },
        }

        return task

    def _add_authorization(self, task):
        if self._access_token:
            task["app_engine_http_request"]["headers"]['X-Forwarded-Authorization'] = f'Bearer {self._access_token}'
        return task

    def _add_payload(self, task):
        if self._payload is not None:
            if not isinstance(self._payload, dict):
                raise TypeError('Payload should be dict type')
            payload = json.dumps(self._payload, ensure_ascii=False).encode('utf-8')
            task["app_engine_http_request"]["body"] = payload
            task["app_engine_http_request"]["headers"]['Content-Type'] = 'application/json'
        return task

    def _convert_in_seconds(self, task):
        if self._in_seconds is not None:
            d = datetime.datetime.utcnow() + datetime.timedelta(seconds=self._in_seconds)
            timestamp = timestamp_pb2.Timestamp()
            timestamp.FromDatetime(d)
            task['schedule_time'] = timestamp
        return task

    @staticmethod
    def _remove_empty_headers(task):
        if len(task['app_engine_http_request']['headers']) == 0:
            del task['app_engine_http_request']['headers']
        return task

    @property
    def _http_method(self):
        if self._update:
            return tasks_v2.HttpMethod.PUT
        else:
            return tasks_v2.HttpMethod.GET if self._payload is None else tasks_v2.HttpMethod.POST


class Storage:
    """ Google Cloud Storage Adapter"""

    def __init__(self, bucket_name='expressmoney'):
        self._bucket_name = bucket_name
        self.storage_client = storage.Client()
        self.bucket = self.storage_client.bucket(self._bucket_name)

    def upload_blob(self, source_file_name, destination_blob_name, type_file='pdf'):
        """Uploads a file to the bucket."""
        blob = self.bucket.blob(destination_blob_name)
        blob.upload_from_string(source_file_name, content_type=f'application/{type_file}')

    def delete_blob(self, blob_name):
        """Deletes a blob from the bucket."""
        blob = self.bucket.blob(blob_name)
        blob.delete()

    def get_blob(self, path_to_file):
        """ Upload a file from the bucket. """
        blob = self.bucket.blob(path_to_file)
        try:
            file = io.BytesIO(blob.download_as_bytes())
            return file
        except NotFound:
            return None


class VertexAI:
    """Google Cloud VertexAI Adapter"""

    REGION = 'us-central1'
    PROJECT_ID = '1086735462412'

    def __init__(self, endpoint_id, instance_dict: dict, action='predict'):
        self._action = action
        self.endpoint_id = endpoint_id
        self.url = f'https://{self.REGION}-aiplatform.googleapis.com/v1/projects/{self.PROJECT_ID}/locations/' \
                   f'{self.REGION}/endpoints/{self.endpoint_id}:{action}'
        self.bearer = self._get_token()
        self._instance_dict = instance_dict
        self.__response = None

    @property
    def _response(self):

        if self.__response is None:
            # preparing data for google service
            data = dict()
            data['instances'] = []
            data['instances'].append(self._instance_dict)

            headers = {
                'Content-Type': 'application/json',
                'Authorization': f'Bearer {self.bearer}',
            }

            # self.__response = RequestService.request(method='POST', data=data, headers=headers, url=self.url)

            self.__response = requests.post(self.url, json=data, headers=headers)

            if self.__response.status_code != 200:
                error = f'Не удалось получить prediction из VertexAI: {self.__response.status_code}'
                report.report(error)
                raise Exception(error)

        return self.__response

    def get_score(self):
        predictions = self._response
        predictions = predictions.json().get('predictions')[0]
        for target_index, target_value in enumerate(predictions['classes']):
            if target_value == '1':
                score = predictions.get('scores')[target_index]
                return score

        raise Exception("Not found target '1' in prediction result")

    def get_explanations(self):
        """Получить весовые значения фичей"""
        if self._action != 'explain':
            return None
        target_name = self._response.json().get('explanations')[0].get('attributions')[0].get('outputDisplayName')
        explanations = self._response.json().get('explanations')[0].get('attributions')[0].get('featureAttributions')
        multiplier = 1000 if int(target_name) == 1 else -1000
        explanations = {key: multiplier * value for key, value in explanations.items()}
        return explanations

    @staticmethod
    def _get_token():
        if os.getenv('GAE_APPLICATION'):
            # when we are in GAE env:
            metadata_url = 'http://metadata.google.internal/computeMetadata/v1/'
            metadata_headers = {'Metadata-Flavor': 'Google'}
            service_account = 'default'

            url = '{}instance/service-accounts/{}/token'.format(
                metadata_url, service_account)

            # Request an access token from the metadata server.
            r = requests.get(url, headers=metadata_headers)
            r.raise_for_status()

            # Extract the access token from the response.
            access_token = r.json()['access_token']
        else:
            # when running locally:
            stream = os.popen('gcloud auth application-default print-access-token')
            access_token = stream.read().strip()

        return access_token


class Request:
    """Sync http request"""

    requests = requests

    def __init__(self, service: str = None, path: str = '/', access_token: str = None, project: str = 'expressmoney',
                 timeout: tuple = (30, 30)):
        self.__path = path
        self.__service = service
        self.__access_token = str(access_token)
        self.__project = project
        self.__timeout = timeout

    def run(self, payload: dict = None, update=False):
        if update:
            response = requests.put(self._uri, json=payload if payload else {}, headers=self._headers,
                                    timeout=self.__timeout)
        elif payload:
            response = requests.post(self._uri, json=payload, headers=self._headers, timeout=self.__timeout)
        else:
            response = requests.get(self._uri, headers=self._headers, timeout=self.__timeout)
        if response.status_code == 403:
            raise Exception('Service account does not have permission '
                            'to access the IAP-protected application.')
        elif response.status_code not in (200, 201):
            raise Exception('Bad response from application: {!r} / {!r} / {!r}'.format(
                response.status_code, response.headers, response.text))
        return response

    @property
    def _uri(self):
        local_url = 'http://127.0.0.1:8000'
        url = f'https://{self.__service}-dot-{self.__project}.appspot.com' if self.__service else local_url
        return f'{url}{self.__path}'

    @property
    def _headers(self):
        headers = dict()
        headers.update(self._get_authorization())
        return headers

    def _get_authorization(self) -> dict:
        return {'X-Forwarded-Authorization': f'Bearer {self.__access_token}'} if self.__access_token else {}



from google.cloud import error_reporting

report = error_reporting.Client()


def report_exception(func):
    """Отправляет ошибку в Error Reporting вместо исключения"""
    def exception_wrapper(**kwargs):
        try:
            result = func(**kwargs)
            return result
        except Exception as exc:
            report.report(str(exc)[:2048])
    return exception_wrapper



from django.conf import settings
from expressmoney.utils import get_ip


def allowed_ip(request):
    user_ip = get_ip(request)

    for allow_ip in settings.ALLOWED_IP:
        if user_ip == allow_ip or user_ip.startswith(allow_ip):
            return True
    return False


def fix_flow_signal(func):
    def _wrapper_fix_signal(self, **kwargs):
        kwargs.update({'self': self})
        func(**kwargs)

    return _wrapper_fix_signal



import json
from typing import Union


def get_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR', None)
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR', None)
    return ip


def get_http_referer(request):
    http_referer = request.META.get('HTTP_REFERER')
    http_host = request.META.get('HTTP_HOST')
    return http_referer if http_referer else http_host


def write_cloud_functions_log(message: Union[str, dict], severity: str = 'DEBUG', request_=None,
                              project: str = 'expressmoney'):
    """ Запись структурированных логов в Cloud Functions logger """

    assert severity in ('DEFAULT', 'NOTICE', 'INFO', 'DEBUG', 'WARNING', 'ERROR', 'CRITICAL', 'ALERT', 'EMERGENCY')

    # Build structured log messages as an object.
    global_log_fields = {}

    # Add log correlation to nest all log messages.
    if request_:
        trace_header = request_.headers.get("X-Cloud-Trace-Context")

        if trace_header and project:
            trace = trace_header.split("/")
            global_log_fields[
                "logging.googleapis.com/trace"
            ] = f"projects/{project}/traces/{trace[0]}"

    # Complete a structured log entry.
    entry = dict(
        severity=severity,
        message=message,
        **global_log_fields,
    )

    print(json.dumps(entry))



from typing import Union

from django.conf import settings
from django.contrib.auth import get_user_model
from rest_framework_simplejwt.tokens import RefreshToken

from expressmoney import adapters
from expressmoney.adapters import Request

User = get_user_model()


class DjangoTasks(adapters.Tasks):

    def __init__(self, user: Union[int, User], queue: str = 'attempts-1', location: str = 'europe-west1',
                 in_seconds: int = None):
        if not isinstance(user, User):
            user = User.objects.get(pk=user)
        access_token = RefreshToken.for_user(user).access_token
        project = settings.PROJECT
        super().__init__(access_token, queue, location, project, in_seconds)


class DjangoPubSub(adapters.PubSub):

    def __init__(self, topic_id: str, user: Union[None, int, User] = None):
        if user:
            if not isinstance(user, User):
                user = User.objects.get(pk=user)
            access_token = RefreshToken.for_user(user).access_token
        else:
            access_token = None
        project = settings.PROJECT
        super().__init__(topic_id, access_token, project)


class DjangoStorage(adapters.Storage):
    def __init__(self):
        bucket_name = f'{settings.PROJECT}.appspot.com'
        super().__init__(bucket_name)


class DjangoRequest(Request):

    def __init__(self, service: str = None, path: str = '/', user: Union[int, User] = None):
        user = user if isinstance(user, User) else User.objects.get(pk=user)
        access_token = RefreshToken.for_user(user).access_token if user else None
        project = settings.PROJECT
        super().__init__(service, path, access_token, project)

    def _get_authorization(self) -> dict:
        from google.auth.transport.requests import Request
        from google.oauth2 import id_token
        authorization = super()._get_authorization()
        open_id_connect_token = id_token.fetch_id_token(Request(), settings.IAP_CLIENT_ID)
        iap_token = {'Authorization': f'Bearer {open_id_connect_token}'}
        authorization.update(iap_token)
        return authorization



"""
py setup.py sdist
twine upload dist/expressmoney-1.0.7.tar.gz
"""
import setuptools

setuptools.setup(
    name='expressmoney',
    packages=setuptools.find_packages(),
    version='1.0.7',
    description='SDK ExpressMoney',
    author='Development team',
    author_email='dev@expressmoney.com',
    # url='https://github.com/zsergey85/',
    install_requires=('google-cloud-secret-manager', 'google-cloud-error-reporting', 'google-cloud-pubsub',
                      'google-cloud-tasks', 'google-cloud-storage', 'requests'),
    # download_url='https://github.com/zsergey85/',
    keywords=('expressmoney',),
    classifiers=(),
    python_requires='>=3.7',
)



from rest_framework import status
from rest_framework.exceptions import APIException


class UnknownException(APIException):
    pass


class ServiceUnavailable(UnknownException):
    status_code = status.HTTP_503_SERVICE_UNAVAILABLE
    default_detail = 'Remote Service temporarily unavailable, try again later.'
    default_code = 'service_unavailable'


class ServiceError(UnknownException):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    default_detail = 'Remote Service API has changed'
    default_code = 'service_api_error'
